;-----------------------------------------------------------------------------------
; BareNumbersOS v.0.3 - GPL3 - Aug 2012
;
;-----------------------------------------------------------------------------------
; Interrupts
;-----------------------------------------------------------------------------------


exc_string db 'Unknown Fatal Exception!', 0
; -----------------------------------------------------------------------------
; Default exception handler
align 16
exception_gate:
	push rsi
	; print something....
	mov rsi, exc_string
	call os_print_string
	pop rsi
	jmp $				; Hang
; -----------------------------------------------------------------------------


; -----------------------------------------------------------------------------
; Default interrupt handler
align 16
interrupt_gate:				; handler for all other interrupts
	iretq				; It was an undefined interrupt so return to caller
; -----------------------------------------------------------------------------


; -----------------------------------------------------------------------------
; Cascade interrupt. IRQ 2, INT 0x22
align 16
cascade:
	push rax

	mov al, 0x20			; Acknowledge the IRQ
	out 0x20, al

	pop rax
	iretq
; -----------------------------------------------------------------------------


; -----------------------------------------------------------------------------
; Timer. IRQ 0 - 100Hz
align 16
clock:
	cld				; Clear direction flag
	add qword [os_Ticks], 1		; 64-bit counter
	;inc	byte [0B8000h+2*0]	; debug: make the ticks appear
	push	rax

	mov	al, 0x20		; Acknowledge the IRQ
	out	0x20, al

	pop	rax
	iretq
; -----------------------------------------------------------------------------


; -----------------------------------------------------------------------------
; Real-time clock interrupt. IRQ 8, INT 0x28
; Currently this IRQ runs 8 times per second (As defined in kinit.inc)
; The supervisor lives here
align 16
rtc:
	push rax
	push rcx
	push rsi
	push rdi

	cld				; Clear direction flag
	add qword [os_ClockCounter], 1	; 64-bit counter started at bootup

	
	; Check to make sure that at least one core is running something
	cmp word [os_QueueLen], 0	; Check the length of the Queue
	jne rtc_end			; If it is greater than 0 then skip to the end

	mov rcx, 256
	mov rsi, cpustatus
nextcpu:
	lodsb
	dec rcx
	bt ax, 1			; Is bit 1 set? If so then the CPU is running a job
	jc rtc_end
	cmp rcx, 0
	jne nextcpu
	mov rax, os_command_line	; If nothing is running then restart the CLI
	call os_smp_enqueue

rtc_end:
	mov al, 0x0C			; Select RTC register C
	out 0x70, al			; Port 0x70 is the RTC index, and 0x71 is the RTC data
	in al, 0x71			; Read the value in register C

	mov al, 0x20			; Acknowledge the IRQ on the PICs
	out 0xA0, al
	out 0x20, al

	pop rdi
	pop rsi
	pop rcx
	pop rax
	iretq
; -----------------------------------------------------------------------------


; -----------------------------------------------------------------------------
; A simple interrupt that just acknowledges an IPI. Useful for getting an AP past a 'hlt' in the code.
align 16
ap_wakeup:
	push rdi
	push rax

	cld				; Clear direction flag
	mov rdi, [os_LocalAPICAddress]	; Acknowledge the IPI
	add rdi, 0xB0
	xor rax, rax
	stosd

	pop rax
	pop rdi
	iretq				; Return from the IPI.
; -----------------------------------------------------------------------------


; -----------------------------------------------------------------------------
; Resets a CPU to execute ap_clear
align 16
ap_reset:
	cld				; Clear direction flag
	mov rax, ap_clear		; Set RAX to the address of ap_clear
	mov [rsp], rax			; Overwrite the return address on the CPU's stack
	mov rdi, [os_LocalAPICAddress]	; Acknowledge the IPI
	add rdi, 0xB0
	xor rax, rax
	stosd
	iretq				; Return from the IPI. CPU will execute code at ap_clear
; -----------------------------------------------------------------------------


; -----------------------------------------------------------------------------
; Enable an interrupt line
; Expects the interrupt line # in AL
align 16
interrupt_enable:
	push rdx
	push rcx
	push rbx
	push rax
	push rax

	in al, 0x21				; low byte target 0x21
	mov bl, al
	pop rax
	mov dx, 0x21				; Use the low byte pic
	cmp al, 8
	jb interrupt_enable_low
	sub al, 8				; IRQ 8-16
	push rax
	in al, 0xA1				; High byte target 0xA1
	mov bl, al
	pop rax
	mov dx, 0xA1				; Use the high byte pic
interrupt_enable_low:
	mov cl, al
	mov al, 1
	shl al, cl
	not al
	and al, bl
	out dx, al

	pop rax
	pop rbx
	pop rcx
	pop rdx
	ret
; -----------------------------------------------------------------------------


; -----------------------------------------------------------------------------
; CPU Exception Gates
align 16
exception_gate_00:
	push rax
	mov al, 0x00
	jmp exception_gate_main

align 16
exception_gate_01:
	push rax
	mov al, 0x01
	jmp exception_gate_main

align 16
exception_gate_02:
	push rax
	mov al, 0x02
	jmp exception_gate_main

align 16
exception_gate_03:
	push rax
	mov al, 0x03
	jmp exception_gate_main

align 16
exception_gate_04:
	push rax
	mov al, 0x04
	jmp exception_gate_main

align 16
exception_gate_05:
	push rax
	mov al, 0x05
	jmp exception_gate_main

align 16
exception_gate_06:
	push rax
	mov al, 0x06
	jmp exception_gate_main

align 16
exception_gate_07:
	push rax
	mov al, 0x07
	jmp exception_gate_main

align 16
exception_gate_08:
	push rax
	mov al, 0x08
	jmp exception_gate_main

align 16
exception_gate_09:
	push rax
	mov al, 0x09
	jmp exception_gate_main

align 16
exception_gate_10:
	push rax
	mov al, 0x0A
	jmp exception_gate_main

align 16
exception_gate_11:
	push rax
	mov al, 0x0B
	jmp exception_gate_main

align 16
exception_gate_12:
	push rax
	mov al, 0x0C
	jmp exception_gate_main

align 16
exception_gate_13:
	push rax
	mov al, 0x0D
	jmp exception_gate_main

align 16
exception_gate_14:
	push rax
	mov al, 0x0E
	jmp exception_gate_main

align 16
exception_gate_15:
	push rax
	mov al, 0x0F
	jmp exception_gate_main

align 16
exception_gate_16:
	push rax
	mov al, 0x10
	jmp exception_gate_main

align 16
exception_gate_17:
	push rax
	mov al, 0x11
	jmp exception_gate_main

align 16
exception_gate_18:
	push rax
	mov al, 0x12
	jmp exception_gate_main

align 16
exception_gate_19:
	push rax
	mov al, 0x13
	jmp exception_gate_main

;-------
align 16
exception_gate_main:
;	push rsi
;	push rdi
;;	call os_smp_get_id	; debug
;	mov rdi, os_temp_string
;	mov rsi, rdi
;	call os_int_to_string
;	call os_print_string
;	pop rdi
;	pop rsi
	pop rax

	call os_speaker_beep
	call os_speaker_beep
	call os_speaker_beep
			
	call init_memory_map
	jmp ap_clear			; jump to AP clear code



; =============================================================================
; EOF
